import traceback
from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
from chellow.models import (
    Session, Contract, Era, Site, SiteEra, Tpr, MeasurementRequirement, Ssc)
from chellow.utils import (
    hh_min, hh_max, req_date, req_int, csv_make_val, c_months_u, to_ct)
from chellow.computer import contract_func, SupplySource
from chellow.views import chellow_redirect
import chellow.computer
import chellow.dloads
from werkzeug.exceptions import BadRequest
import csv
import os
import threading
from flask import g


def create_csv(f, sess, start_date, finish_date, contract_id):
    caches = {}
    writer = csv.writer(f, lineterminator='\n')
    contract = Contract.get_supplier_by_id(sess, contract_id)
    forecast_date = chellow.computer.forecast_date()

    start_date_ct, finish_date_ct = to_ct(start_date), to_ct(finish_date)

    month_pairs = c_months_u(
        start_year=start_date_ct.year, start_month=start_date_ct.month,
        finish_year=finish_date_ct.year, finish_month=finish_date_ct.month)

    bill_titles = contract_func(caches, contract, 'virtual_bill_titles')()

    for tpr in sess.query(Tpr).join(MeasurementRequirement).join(Ssc). \
            join(Era).filter(
                Era.start_date <= finish_date, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= start_date), or_(
                    Era.imp_supplier_contract == contract,
                    Era.exp_supplier_contract == contract)
            ).order_by(Tpr.code).distinct():
        for suffix in ('-kwh', '-rate', '-gbp'):
            bill_titles.append(tpr.code + suffix)
    writer.writerow(
        ['MPAN Core', 'Site Code', 'Site Name', 'Account', 'From', 'To'] +
        bill_titles)
    vb_func = contract_func(caches, contract, 'virtual_bill')

    for month_start, month_finish in month_pairs:
        period_start = hh_max(start_date, month_start)
        period_finish = hh_min(finish_date, month_finish)

        for era in sess.query(Era).distinct().filter(
                or_(
                    Era.imp_supplier_contract == contract,
                    Era.exp_supplier_contract == contract),
                Era.start_date <= period_finish,
                or_(
                    Era.finish_date == null(),
                    Era.finish_date >= period_start)):

            chunk_start = hh_max(period_start, era.start_date)
            chunk_finish = hh_min(period_finish, era.finish_date)

            polarities = []
            if era.imp_supplier_contract == contract:
                polarities.append(True)
            if era.exp_supplier_contract == contract:
                polarities.append(False)
            for polarity in polarities:
                vals = []
                data_source = SupplySource(
                    sess, chunk_start, chunk_finish, forecast_date, era,
                    polarity, caches)

                site = sess.query(Site).join(SiteEra).filter(
                    SiteEra.era == era, SiteEra.is_physical == true()).one()

                vals = [
                    data_source.mpan_core, site.code, site.name,
                    data_source.supplier_account, data_source.start_date,
                    data_source.finish_date
                ]

                vb_func(data_source)
                bill = data_source.supplier_bill
                for title in bill_titles:
                    if title in bill:
                        val = bill[title]
                        del bill[title]
                    else:
                        val = ''
                    vals.append(val)

                for k in sorted(bill.keys()):
                    vals.append(k)
                    vals.append(str(bill[k]))
                writer.writerow(csv_make_val(v) for v in vals)


def content(start_date, finish_date, contract_id, user):
    sess = f = None
    try:
        running_name, finished_name = chellow.dloads.make_names(
            'virtual_bills.csv', user)

        f = open(running_name, mode='w', newline='')
        sess = Session()

        create_csv(f, sess, start_date, finish_date, contract_id)

    except BadRequest as e:
        if f is None:
            raise
        else:
            f.write('\n' + e.description)
    except BaseException:
        if f is None:
            raise
        else:
            f.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date('start')
    finish_date = req_date('finish')
    contract_id = req_int('supplier_contract_id')

    threading.Thread(
        target=content,
        args=(start_date, finish_date, contract_id, g.user)).start()
    return chellow_redirect("/downloads", 303)



from datetime import datetime as Datetime
from dateutil.relativedelta import relativedelta
from flask import render_template


def do_get(sess):
    init = Datetime.utcnow()
    init = Datetime(init.year, init.month, 1) - relativedelta(months=1)
    return render_template('report_155.html', init=init)



import traceback
from sqlalchemy.sql import func
from chellow.models import (
    Batch, Session, Contract, Bill, GBatch, GContract, GBill)
from chellow.views import chellow_redirect
import chellow.dloads
import csv
from flask import g
import threading
import os
from chellow.utils import csv_make_val


def content(user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'batches.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        titles = (
            'utility', "chellow_id", "reference", "description",
            "contract_name", "num_bills", "net_gbp", "vat_gbp", "gross_gbp",
            "kwh"
        )
        writer.writerow(titles)

        for batch, contract in sess.query(Batch, Contract).join(
                Contract).order_by(Batch.contract_id, Batch.reference):

            (
                num_bills, sum_net_gbp, sum_vat_gbp, sum_gross_gbp,
                sum_kwh) = sess.query(
                func.count(Bill.id), func.sum(Bill.net), func.sum(Bill.vat),
                func.sum(Bill.gross), func.sum(Bill.kwh)).filter(
                Bill.batch == batch).one()

            if sum_net_gbp is None:
                sum_net_gbp = sum_vat_gbp = sum_gross_gbp = sum_kwh = 0
            vals = {
                'utility': 'electricity',
                'chellow_id': batch.id,
                'reference': batch.reference,
                'description': batch.description,
                'contract_name': contract.name,
                'num_bills': num_bills,
                'net_gbp': sum_net_gbp,
                'vat_gbp': sum_vat_gbp,
                'gross_gbp': sum_gross_gbp,
                'kwh': sum_kwh
            }

            writer.writerow(csv_make_val(vals[t]) for t in titles)

            # Avoid a long-running transaction
            sess.rollback()

        for g_batch, g_contract in sess.query(GBatch, GContract).join(
                GContract).order_by(GBatch.g_contract_id, GBatch.reference):

            (
                num_bills, sum_net_gbp, sum_vat_gbp, sum_gross_gbp,
                sum_kwh) = sess.query(
                func.count(GBill.id), func.sum(GBill.net), func.sum(GBill.vat),
                func.sum(GBill.gross), func.sum(GBill.kwh)).filter(
                GBill.g_batch == g_batch).one()

            if sum_net_gbp is None:
                sum_net_gbp = sum_vat_gbp = sum_gross_gbp = sum_kwh = 0
            vals = {
                'utility': 'gas',
                'chellow_id': g_batch.id,
                'reference': g_batch.reference,
                'description': g_batch.description,
                'contract_name': g_contract.name,
                'num_bills': num_bills,
                'net_gbp': sum_net_gbp,
                'vat_gbp': sum_vat_gbp,
                'gross_gbp': sum_gross_gbp,
                'kwh': sum_kwh
            }

            writer.writerow(csv_make_val(vals[t]) for t in titles)

            # Avoid a long-running transaction
            sess.rollback()

    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    args = (g.user,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import collections
import datetime
import importlib
import threading
import traceback
from pkgutil import iter_modules

import chellow
from chellow.models import (
    BillType, GBatch, GReadType, GSupply, GUnit, Session,
)

import pytz

from werkzeug.exceptions import BadRequest

importer_id = 0
import_lock = threading.Lock()
importers = {}


def find_parser_names():
    return ', '.join(
        '.' + name[14:].replace('_', '.') for module_finder, name, ispkg in
        iter_modules(chellow.__path__) if name.startswith('g_bill_parser_'))


class GBillImporter(threading.Thread):
    def __init__(self, sess, g_batch_id, file_name, file_bytes):
        threading.Thread.__init__(self)
        global importer_id
        self.importer_id = importer_id
        importer_id += 1

        self.g_batch_id = g_batch_id
        if len(file_bytes) == 0:
            raise BadRequest("File has zero length")
        imp_mod = None
        parts = file_name.split('.')[::-1]
        for i in range(len(parts)):
            nm = 'g_bill_parser_' + '_'.join(parts[:i+1][::-1]).lower()
            try:
                imp_mod = nm, importlib.import_module('chellow.' + nm)
            except ImportError:
                pass

        if imp_mod is None:
            raise BadRequest(
                "Can't find a parser for the file '" + file_name +
                "'. The file name needs to have an extension that's one of " +
                "the following: " + find_parser_names() + ".")

        self.parser_name = imp_mod[0]
        self.parser = imp_mod[1].Parser(file_bytes)
        self.successful_bills = []
        self.failed_bills = []
        self.log = collections.deque()
        self.bill_num = None

    def _log(self, msg):
        with import_lock:
            self.log.appendleft(
                datetime.datetime.now(pytz.utc).strftime("%Y-%m-%d %H:%M:%S") +
                ' - ' + msg)

    def status(self):
        return "Parsed file up to line number: " + \
            str(self.parser.line_number) + \
            ". Inserting raw bills up to bill: " + str(self.bill_num) + "."

    def run(self):
        sess = None
        try:
            self._log(
                "Starting to parse the file with '" + self.parser_name + "'.")
            sess = Session()
            g_batch = GBatch.get_by_id(sess, self.g_batch_id)
            raw_bills = self.parser.make_raw_bills()
            self._log(
                "Successfully parsed the file, and now I'm starting to "
                "insert the raw bills.")
            for self.bill_num, raw_bill in enumerate(raw_bills):
                try:
                    bill_type = BillType.get_by_code(
                        sess, raw_bill['bill_type_code'])
                    g_supply = GSupply.get_by_mprn(sess, raw_bill['mprn'])
                    g_bill = g_batch.insert_g_bill(
                        sess, g_supply, bill_type,
                        raw_bill['reference'], raw_bill['account'],
                        raw_bill['issue_date'], raw_bill['start_date'],
                        raw_bill['finish_date'], raw_bill['kwh'],
                        raw_bill['net_gbp'], raw_bill['vat_gbp'],
                        raw_bill['gross_gbp'], raw_bill['raw_lines'],
                        raw_bill['breakdown'])
                    sess.flush()
                    for raw_read in raw_bill['reads']:
                        prev_type = GReadType.get_by_code(
                            sess, raw_read['prev_type_code'])
                        pres_type = GReadType.get_by_code(
                            sess, raw_read['pres_type_code'])
                        g_unit = GUnit.get_by_code(sess, raw_read['unit'])
                        g_read = g_bill.insert_g_read(
                            sess, raw_read['msn'], g_unit,
                            raw_read['correction_factor'],
                            raw_read['calorific_value'],
                            raw_read['prev_value'], raw_read['prev_date'],
                            prev_type, raw_read['pres_value'],
                            raw_read['pres_date'], pres_type)

                        sess.expunge(g_read)
                    self.successful_bills.append(raw_bill)
                    sess.expunge(g_bill)
                except BadRequest as e:
                    sess.rollback()
                    raw_bill['error'] = e.description
                    self.failed_bills.append(raw_bill)

            if len(self.failed_bills) == 0:
                sess.commit()
                self._log(
                    "All the bills have been successfully loaded and attached "
                    "to the batch.")
            else:
                sess.rollback()
                self._log(
                    "The import has finished, but " +
                    str(len(self.failed_bills)) + " bills failed to load, " +
                    "and so the whole import has been rolled back.")

        except BadRequest as e:
            self._log(e.description)
        except BaseException:
            self._log("I've encountered a problem: " + traceback.format_exc())
        finally:
            if sess is not None:
                sess.rollback()
                sess.close()

    def make_fields(self):
        with import_lock:
            fields = {
                'log': tuple(self.log), 'is_alive': self.isAlive(),
                'importer_id': self.importer_id}
            if not self.isAlive():
                fields['successful_bills'] = self.successful_bills
                fields['failed_bills'] = self.failed_bills
            return fields


def start_bill_importer(sess, batch_id, file_name, file_bytes):
    with import_lock:
        bi = GBillImporter(sess, batch_id, file_name, file_bytes)
        importers[bi.importer_id] = bi
        bi.start()

    return bi.importer_id


def get_bill_importer_ids(g_batch_id):
    with import_lock:
        return [
            k for k, v in importers.items() if v.g_batch_id == g_batch_id]


def get_bill_importer(id):
    with import_lock:
        return importers[id]
